#!/usr/bin/env python

# undistract - monitor work habits and nudge the user to get back to work
# (c) 2006-2007 Andy O'Neill <andy@potatoriot.com>

# Requirements: 
#   1) PyXss X-Screensaver Extension for Python
#           http://bebop.bigasterisk.com/python/
#           (in Ubuntu, package x11proto-scrnsaver-dev is required to build 
#           that module (get xlibs-dev, libxss-dev too, and maybe some other
#           stuff))
#   2) Python bindings to libwnck, for window manager interaction
#           (in Ubuntu, install python-gnome2-desktop)
#   3) Python bindings to libnotify, package 'python-notify' in Ubuntu
#   4) Standard *nix command: 'ping'


# Configuration
#   defaults in this file
#   user settings in ~/.undistract

#   LICENSE: GPL
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Library General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import datetime, time
import ConfigParser
import os, sys, re
from locale import atof, atoi
# linwnck python bindings
import wnck, gtk, gobject
import xss
import pynotify

# some global runtime variables
hosts = []
distractions = []
distraction_level = 0
last_warning = -1
last_minimize = -1

############################# HOST CLASS #####################################
class Host:
    """A class useful for determing if a given host is on the local network,
    which I use to determine if I am 'at work' or not"""
    def __init__(self, hostname):
        self.name = hostname  # fqdn of host
        ##################    CONFIG PARAMETERS   ##########################
        self.bad_ping_interval = 15*60 # don't retry ping for some time if 
                                       # it fails (e.g network is down)
                                       # networks don't change very quickly!
        self.max_ping_time = 10  # in secs, assume host is local if ping time is
                                 # less than this
        self.ping_interval = 300 # interval with which to ping (s)


        ##################    RUNTIME DATA PARAMETERS   #######################
        self.next_ping = -1      # time (epoch) of next ping
        self.last_bad_ping = -1  # store time of previous ping failure
        self.present = False     # stored result for between ping intervals

    def is_present(self):
        """Return true if we are on the same network as this host"""
        now = time.time()
        # check to see if there was a bad ping recently, if so we won't bother
        if self.last_bad_ping > 0 and (now - self.last_bad_ping) < \
            self.bad_ping_interval:
            self.present = False
            debug(self.name + ": Last ping failed, not attempting ping for " + \
                str(self.last_bad_ping + self.bad_ping_interval - now), level=3)
            return self.present

        # check to see if we recently did a successful ping
        if self.next_ping > 0 and self.next_ping > now:
            # use stored value
            debug(self.name + ": Using stored ping status: " + \
                str(self.present), level=3)
            return self.present
        else:
            ping_cmd = "ping -c 1 -W 1 "
            self.next_ping = now + self.ping_interval
            r = re.compile(r"[= ]")
            ping = os.popen(ping_cmd + self.name).readlines()
            # check for bad/missing hosts
            if len(ping) == 0:
                self.present = False
                self.last_bad_ping = now
                debug(self.name + ": Disabling since unknown host in ping" )
                return self.present
            if "0 received" in ping[-1]: 
                self.present = False
                self.last_bad_ping = now
                debug(self.name + ": Disabling since not responding to ping" )
                return self.present

            # examine ping results
            self.present = False
            for line in ping:
                if "time=" in line:
                    ping_time = atof(r.split(line)[-2])
                    if ping_time < self.max_ping_time:
                        self.present = True
                        debug("Host %s is local: ping time was %f" % \
                                    ( self.name, ping_time ), level=2 )
                        return self.present
                    else:
                        self.present = False
                        debug("Host %s is not local: ping time was %f" % \
                                    ( self.name, ping_time ), level=2 )
                        return self.present
            
            self.present = False
            debug("Host %s is not responding to ping" % \
                        ( self.name ), level=2 )
            return self.present
        
############################ CONFIG CLASS #####################################
class Config:
    """Class to read and store config parameters"""
    def __init__(self, config_file="~/.undistract"):
        self.debug = 1           # debug level: 0=off, 1=some,... 4=TMI!
        self.work_days = [ 1, 2, 3, 4, 5 ]  # days of week that we ought to be
                                            # working not distracted
                                            # 1 = monday, 0 = sunday
        self.work_start_time = "10:00"  # hour at which we have to start work
        self.work_end_time = "19:00"    # hour at which we can start slacking
        self.work_hosts = []            # names of hosts on the work network
        # names of programs that are distracting
        self.distractions = ["firefox", "firefox-bin", "konqueror", "opera" ]

        self.warn_level = 2250          # when the 'distraction_level' reaches
                                        # this, the user will be chastised
        self.act_level = 5000           # when the distraction_level reaches
                                        # this, the user's windows will be 
                                        # minimized
        self.max_distraction_level = 10000 # ceiling for distraction_level
        self.idle_time = 10*60          # time (s) to consider session idle
        self.distracted_penalty = 10    # amount to increase distraction_level
                                        # by if the user is distracted
        self.working_reward = 20        # amount to decrease level by if the
                                        # user is not distracted or idle
        self.idle_reward = 10           # amount to reduce level by if the user
                                        # is idle
        self.log_filename = ""          # log to this file
        self.minimize_delay = 5000      # length of time while we minimize
        self.warn_interval = 60         # minumum interval between warnings
        self.minimize_interval = 20     # minimum interval between actions

        # handy runtime data variable
        self.lf = sys.stderr 
        self.config_file = os.path.expanduser(config_file)

        # read settings from the config file
        if os.path.exists(self.config_file):
            self.read_file()

        # open the log file if one has been set
        if self.log_filename != "":
            print "Redirecting log output to " + self.log_filename
            try:
                self.lf = open(self.log_filename, 'a')
            except:
                print "Error opening log file " + self.log_filename
                self.lf = sys.stderr

    def read_file(self):
        """Read settings from the config file"""
        print "Reading config file: " + self.config_file
        parser = ConfigParser.SafeConfigParser()
        parser.read([os.path.expanduser(self.config_file)])
        if parser.has_option("default", "work_host"):
            for hostname in parser.get("default", "work_host").split():
                self.work_hosts.append(hostname)
        if parser.has_option("default", "distractions"):
            self.distractions = []
            for program in parser.get("default", "distractions").split():
                self.distractions.append(program)
        if parser.has_option("default", "work_time"):
            work_time_string = parser.get("default", "work_time")
            r = re.compile(r"[:-]")
            split = r.split(work_time_string)
            self.work_start_time = datetime.time(atoi(split[0]), 
                    atoi(split[1]))
            self.work_end_time = datetime.time(atoi(split[2]), 
                    atoi(split[3]))
        if parser.has_option("default", "work_days"):
            self.work_days = []
            for day in parser.get("default", "work_days").split():
                self.work_days.append(atoi(day))
        if parser.has_option("default", "log_file"):
            self.log_filename = parser.get("default", "log_file")
            self.log_filename = os.path.expanduser(self.log_filename)
        if parser.has_option("default", "debug"):
            self.debug = parser.getint("default", "debug")
        if parser.has_option("default", "idle_time"):
            self.idle_time = parser.getint("default", "idle_time")
        if parser.has_option("default", "warn_level"):
            self.warn_level = parser.getint("default", "warn_level")
        if parser.has_option("default", "act_level"):
            self.act_level = parser.getint("default", "act_level")
        if parser.has_option("default", "warn_interval"):
            self.warn_interval = parser.getint("default", "warn_interval")
        if parser.has_option("default", "minimize_interval"):
            self.minimize_interval = \
                parser.getint("default", "minimize_interval")
        if parser.has_option("default", "minimize_delay"):
            self.minimize_delay = parser.getint("default", "minimize_delay")
        if parser.has_option("default", "distracted_penalty"):
            self.distracted_penalty = \
                parser.getint("default", "distracted_penalty")
        if parser.has_option("default", "working_reward"):
            self.working_reward = parser.getint("default", "working_reward")
        if parser.has_option("default", "idle_reward"):
            self.idle_reward = parser.getint("default", "idle_reward")

# Global instance of Config class
config = Config()


####################### DISTRACTION CLASS #####################################
class Distraction:
    """Class representing programs which are distracting, e.g. browsers"""
    def __init__(self, name):
        self.name = name

        # runtime variables
        self.became_active = 0
        self.minimize_timeout = None
        self.notification = None
    def is_active(self):
        """Monitor the usage of this application and react accordingly"""
        debug(self.name + ": current level = " + str(distraction_level), 
                level=4)
        active_window = get_active_window()
        if self.name == get_exe(active_window):
            if self.became_active == -1:
                debug("Setting became_active for " + self.name, level=3)
                self.became_active = time.time()
            return True
        else:
            debug("Setting became_active = -1 for " + self.name, level=3)
            self.became_active = -1
            return False
    def active_time(self):
        """Check for activity and then return the time this window has been
        active"""
        self.is_active()
        if self.became_active == -1:
            return 0
        else:
            active_time = time.time() - self.became_active
            debug(self.name + " is the active window, t = %2.0fs" % active_time)
            return active_time
    def get_windows(self):
        windows = []
        for w in get_all_windows():
            if get_exe(w) == self.name:
                windows.append(w)
        return windows

    def start_minimize(self):
        """Start a timeout which will result in the windows being minimized"""
        #for w in self.get_windows():
        #    w.activate(0)
        if self.minimize_timeout != None:
            debug("Not starting minimize threat since one was already started",
                level=2)
            return
        debug("Threatening to minimize", level=2)
        text = "You appear to be distracted.\nI'm going to close your window!"
        n = pynotify.Notification("Undistract", text, "dialog-warning")
        n.set_urgency(pynotify.URGENCY_CRITICAL)
        n.set_timeout(config.minimize_delay)
        n.add_action("clicked","OK! I'll do some work!", notification_cancel_cb,
                self)
        n.add_action("default", "Default Action", notification_cancel_cb, self)
        n.set_hint("x", 0)
        n.set_hint("y", 10)
        n.show()
        self.notification = n
        debug("Initializing timeout for minimize", level=3)
        self.minimize_timeout = gobject.timeout_add(config.minimize_delay,
                self.minimize)
    def minimize(self):
        debug("Minimizing all windows for " + self.name)
        if self.minimize_timeout:
            gobject.source_remove(self.minimize_timeout)
        for w in self.get_windows():
            w.minimize()
        self.minimize_timeout = None
        return False # don't repeat the function
    def cancel_minimize(self):
        debug("Cancelling the minimize timeout", level=2)
        if self.minimize_timeout:
            gobject.source_remove(self.minimize_timeout)
        self.minimize_timeout = None

        
############################ DECISION FUNCTIONS  #############################

def should_be_working():
    """Return true if the user should be working, false if the user can be
    slacking, based on whether the user is at work and what times they should 
    be working:
        * If the user is at work, they should be working
        * If the user is not at work but it's during their work day, they should
          be working
        * Only if the user is not at work and it is outside work hours can they
          slack off in peace
    """
    global config, hosts

    # check to see if we are physically (networkly) AT work
    if len(hosts) > 0:
        debug("Checking network to see if we're at work", level=3)
        for host in hosts:
            # every work host must NOT be present
            if host.is_present():
                # if the user is at work, they should definitely be working,
                # otherwise the could GO HOME!
                return True

    # no work hosts provided or none local, so look at the time to see if we
    # should be working
    if during_work_hours():
        debug("During work hours", level = 3)
        return True
    else:
        debug("Outside of work hours and not at work", level = 4)
        return False

def during_work_hours():
    """Decide if this is a time during which the user should be working.
    Returns True if during work day, False if outside work hours
    Work days are set in config file, work hours are too"""
    global config
    day = time.localtime()[6] + 1
    if day not in config.work_days:
        debug ("Not a work day today (%d)" % day, level=3)
        return False
    
    cur = datetime.time(time.localtime()[3], time.localtime()[4])
    if config.work_start_time < config.work_end_time:
        # assume start time & end_time are on same day
        if cur > config.work_start_time and cur < config.work_end_time:
            return True  # at work
        else:
            return False
    else:
        # assume start_time and end_time are on different days
        if cur > config.work_start_time and cur < datetime.time(23, 59, 59):
            return True
        elif cur > datetime.time(0, 0, 0) and cur < config.work_end_time:
            return True
        else:
            return False

############################# IDLE - XSS INTEGRATION #######################
def is_idle():
    """Check to see if session is currently idle. Returns True for idle"""
    info = xss.get_info()
    if info.idle > config.idle_time * 1000:
        debug("Session is idle for more than " + str(config.idle_time) + "s")
        return True
    else:
        return False


##################### LIBNOTIFY FUNCTIONS #################################
def warn():
    global last_warning
    # check to see how long since last warning
    if last_warning > 0 and (time.time() - last_warning) < config.warn_interval:
        debug("No warning since too soon", level=3)
        return
    text = "You appear to be distracted.\nGet back to work!"
    timeout = 5000
    n = pynotify.Notification("Undistract", text, "dialog-warning")
    n.set_urgency(pynotify.URGENCY_NORMAL)
    n.set_timeout(timeout)
    n.set_hint("x", 0)
    n.set_hint("y", 10)
    n.show()
    last_warning = time.time()
    
def notification_cancel_cb(notification, action, data):
    debug("The notify cancel button was clicked", level=3)
    data.minimize()
    data.cancel_minimize()
    notification.close()


############################# WNCK FUNCTIONS ############################
def get_active_window():
    """Use libwnck to get the active window, return the wnck.Window class"""
    active = None
    windows = get_all_windows()
    for w in windows:
        if w.is_active():
            active = w
    return active

def get_exe(window):
    if window == None:
        return ""
    pid = window.get_pid()
    exe_link = "/proc/" + str(pid) + "/exe"
    exe = ""
    if os.path.exists(exe_link):
        exe = os.path.realpath(exe_link)
    exe_base = os.path.basename(exe)
    debug("Executable of process " + str(pid) + " is " + exe_base, level=4)
    return exe_base

def get_app_windows(window):
    app_windows = []
    exe = get_exe(window)
    for w in get_all_windows():
        if get_exe(w) == exe:
            app_windows.append(w)
    return app_windows


def get_all_windows():
    screen = wnck.screen_get_default()
    active = None
    windows = screen.get_windows()
    if len(windows) == 0:
        screen.force_update()
        windows = screen.get_windows()
    if len(windows) == 0:
        debug("Couldn't get windows of this screen even after update!")
    return windows



################# CONVENIENCE FUNCTIONS   ################################
def debug(msg, level=1):
    """Print out message if config.debug is set"""
    if config.lf == None:
        print "log file is not open"
    if config.debug >= level:
        print >> config.lf, "undistract: " + msg
        config.lf.flush()

def raise_distraction_level(delta = 10):
    """Convenience function to raise distraction level up to a ceiling"""
    global distraction_level
    distraction_level += delta
    if distraction_level > config.max_distraction_level:
        distraction_level = max_distraction_level
    debug("Distraction level is " + str(distraction_level), level=1)

def lower_distraction_level(delta = 10):
    """Convenience function to lower Distraction level but keep it >= 0"""
    global distraction_level
    distraction_level -= delta
    if distraction_level < 0:
        distraction_level = 0
        debug("Distraction level is " + str(distraction_level), level=3)
    else:
        # don't write anything if level = 0... so much boring!
        debug("Distraction level is " + str(distraction_level), level=1)


##################### MAIN FUNCTIONS ####################################

def init():
    global distractions, hosts
    pynotify.init("Undistract")
    # Create 'Host' and 'Distraction' instances
    for h in config.work_hosts:
        hosts.append(Host(h))
    for d in config.distractions:
        distractions.append(Distraction(d))

    if len(distractions) == 0:
        print "No distraction programs were defined"
        sys.exit()

# monitor_all: the repeating function that checks the distractions
def monitor_all(source_id=None):
    """Go through all distractions and see if any of them are being used"""
    global distractions, hosts, last_minimize
    interval = 2500
    active_distraction = None  
    if not should_be_working():
        # increase sleep interval and reset distraction_level to zero
        debug("Should not be working.", level=3)
        interval = 30000
        lower_distraction_level(config.max_distraction_level)
    if is_idle():
        # lower warning level somewhat, sleep for slightly longer
        interval = 10000
        lower_distraction_level(config.idle_reward)
    else:
        active_window = get_active_window()
        if active_window == None:
            debug("No active window")
            # Treat as idle
            interval = 10000
            lower_distraction_level(config.idle_reward)
        else:
            debug("Active program is " + get_exe(active_window), level=3)
            for d in distractions:
                if d.active_time() > 0:
                    active_distraction = d
            if active_distraction:
                raise_distraction_level(config.distracted_penalty)
            else:
                lower_distraction_level(config.working_reward)

    # Act or warn in necessary
    if distraction_level >= config.act_level and active_distraction != None:
        if last_minimize < 0 or \
            time.time() - last_minimize > config.minimize_interval:
            active_distraction.start_minimize()
            last_minimize = time.time()
    elif distraction_level >= config.warn_level and active_distraction != None:
        warn()
    timeout = gobject.timeout_add(interval, monitor_all)
    return False # do not repeatedly call, I have set another timeout

# Initialize and start the program
if __name__ == '__main__':
    init()
    debug("Debug level is " + str(config.debug))
    # Queue up 'monitor_all' to get things started
    gobject.idle_add(monitor_all)
    debug("Entering GTK main loop")
    gtk.main()


